package org.gradoop.famer.linking.similarityMeasuring.methods;

import org.gradoop.common.model.impl.properties.PropertyValue;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import static org.apache.commons.lang.StringUtils.isBlank;

public class TFIDF implements SimilarityComputation<String> {

  /**
   * Tokenizer to split sentences in words
   */
  private final String tokenizer;

  /**
   * Map holding the document frequency of each word
   */
  private final Map<String, Integer> documentFrequency;

  /**
   * @param documentFrequency a mapping of the document frequency to each word
   * @param tokenizer sentence tokenizer
   */
  public TFIDF(final Map<String, Integer> documentFrequency, final String tokenizer) {
    this.documentFrequency = documentFrequency;
    this.tokenizer = tokenizer;
  }

  @Override
  public double computeSimilarity(String atr1, String atr2) throws Exception {
    Objects.requireNonNull(atr1, "First attribute should not be null");
    Objects.requireNonNull(atr2, "Second attribute should not be null");
    if (isBlank(atr1) || isBlank(atr2)) {
      return 0;
    }

    final List<String> doc1Words = Arrays.asList(atr1.split(tokenizer));
    final List<String> doc2Words = Arrays.asList(atr2.split(tokenizer));
    final List<String> terms = extractTerms(doc1Words, doc2Words);
    final Double[] doc1Vector = terms.stream().map(term -> tfidf(term, doc1Words)).toArray(Double[]::new);
    final Double[] doc2Vector = terms.stream().map(term -> tfidf(term, doc2Words)).toArray(Double[]::new);
    return computeCosineSimilarity(doc1Vector, doc2Vector);
  }

  @Override
  public String parsePropertyValue(PropertyValue value) {
    return getString(value);
  }

  /**
   * Puts only unique terms from both docs to a single {@link ArrayList}
   *
   * @param doc1 the first document
   * @param doc2 the second document
   * @return duplicate free List
   */
  private List<String> extractTerms(final List<String> doc1, final List<String> doc2) {
    final Set<String> result = new HashSet<>(doc1);
    result.addAll(doc2);
    return new ArrayList<>(result);
  }

  /**
   * @param term the term
   * @param doc the document
   * @return tfidf score
   */
  private Double tfidf(String term, List<String> doc) {
    return tf(term, doc) * idf(term);
  }

  @Nonnull
  private Double tf(final String term, final List<String> doc) {
    long count = doc.stream().filter(d -> d.equalsIgnoreCase(term)).count();
    return (double) count;
  }

  /**
   * Calculate the inverse document frequency after  log(|D| / 1 + |{d:ti ∈ d}|)
   *
   * @param term the term
   * @return the inverse document frequency
   */
  @Nonnull
  private Double idf(final String term) {
    long df = documentFrequency.get(term);
    if (df == 0) {
      return 0D;
    }
    return Math.log(2 / (float) 1 + df);
  }

  /**
   * Computes the cosine similarity of two vectors
   *
   * @param vector1 the first vector
   * @param vector2 the second vector
   * @return the cosine similarity of the two vectors
   */
  private double computeCosineSimilarity(final Double[] vector1, final Double[] vector2) {
    double ab = 0;
    for (int i = 0; i < vector1.length; i++) {
      ab += vector1[i] * vector2[i];
    }
    double norm1 = computeVectorNorm(vector1);
    double norm2 = computeVectorNorm(vector2);

    return ab / (norm1 * norm2);
  }

  /**
   * Computes the norm of a vector
   *
   * @param vector a vector
   * @return the norm of this vector
   */
  private double computeVectorNorm(final Double[] vector) {
    double sum = 0;
    for (Double aFloat : vector) {
      sum += aFloat * aFloat;
    }
    return Math.sqrt(sum);
  }

}
